"""
Drop-in replacements for `datetime` objects with better typing information.

Note: at runtime, actual `datetime` classes are used directly for minimum
performance overhead.
"""

__all__ = [
    "AwareDateTime",
    "AwareTime",
    "Date",
    "DateTime",
    "NaiveDateTime",
    "NaiveTime",
    "Time",
]

import sys
from datetime import date as _date
from datetime import datetime as _datetime
from datetime import time as _time
from datetime import timedelta as _timedelta
from datetime import tzinfo as _tzinfo
from typing import (
    ClassVar,
    Generic,
    SupportsIndex,
    overload,
)

from typing_extensions import Self, TypeAlias, TypeIs, TypeVar

_MaybeTZ = TypeVar(
    "_MaybeTZ",
    bound=_tzinfo | None,
    default=_tzinfo | None,
    covariant=True,
)
_OptionalTZ = TypeVar("_OptionalTZ", bound=_tzinfo | None)
_TZ = TypeVar("_TZ", bound=_tzinfo)

#
# === Date ===

# subclass _date to allow Date to be used in places where _date is expected
class Date(_date):
    """
    Date object with no time associated.
    """

    # restrict `datetime.date -> Self` to prevent comparisons with `DateTime`
    def __le__(self, value: Self, /) -> bool: ...
    def __lt__(self, value: Self, /) -> bool: ...
    def __ge__(self, value: Self, /) -> bool: ...
    def __gt__(self, value: Self, /) -> bool: ...
    #
    @overload
    def __sub__(self, value: Self, /) -> _timedelta: ...
    @overload
    def __sub__(self, value: _timedelta, /) -> Self: ...

#
# === Time & co. ===

class Time(_time, Generic[_MaybeTZ]):
    """
    Generic Time object with optional timezone attached.
    """

    # replace `datetime.time -> Time`
    min: ClassVar[Time]
    max: ClassVar[Time]

    def __new__(
        cls,
        hour: SupportsIndex = ...,
        minute: SupportsIndex = ...,
        second: SupportsIndex = ...,
        microsecond: SupportsIndex = ...,
        tzinfo: _MaybeTZ = None,
        *,
        fold: int = ...,
    ) -> Time[_MaybeTZ]: ...
    #
    # fix inherited stubs to consider generic timezone
    @property
    def tzinfo(self) -> _MaybeTZ: ...

    if sys.version_info >= (3, 13):
        def __replace__(
            self,
            /,
            *,
            hour: SupportsIndex = ...,
            minute: SupportsIndex = ...,
            second: SupportsIndex = ...,
            microsecond: SupportsIndex = ...,
            tzinfo: _OptionalTZ = ...,
            fold: int = ...,
        ) -> Time[_OptionalTZ]: ...

    def replace(
        self,
        hour: SupportsIndex = ...,
        minute: SupportsIndex = ...,
        second: SupportsIndex = ...,
        microsecond: SupportsIndex = ...,
        tzinfo: _OptionalTZ = None,
        *,
        fold: int = ...,
    ) -> Time[_OptionalTZ]: ...
    #
    # prevent aware-vs-naive comparisons
    @overload
    def __le__(self: NaiveTime, value: NaiveTime, /) -> bool: ...
    @overload
    def __le__(self: AwareTime, value: AwareTime, /) -> bool: ...
    @overload
    def __lt__(self: NaiveTime, value: NaiveTime, /) -> bool: ...
    @overload
    def __lt__(self: AwareTime, value: AwareTime, /) -> bool: ...
    @overload
    def __ge__(self: NaiveTime, value: NaiveTime, /) -> bool: ...
    @overload
    def __ge__(self: AwareTime, value: AwareTime, /) -> bool: ...
    @overload
    def __gt__(self: NaiveTime, value: NaiveTime, /) -> bool: ...
    @overload
    def __gt__(self: AwareTime, value: AwareTime, /) -> bool: ...

NaiveTime: TypeAlias = Time[None]
"""Alias for Time with no timezone associated."""
AwareTime: TypeAlias = Time[_tzinfo]
"""Alias for Time with timezone associated."""

#
# === DateTime & co. ===

class DateTime(_datetime, Generic[_MaybeTZ]):
    """
    Generic DateTime object with optional timezone attached.
    """

    # replace `datetime.date -> DateTime`
    min: ClassVar[DateTime]
    max: ClassVar[DateTime]

    def __new__(
        cls,
        year: SupportsIndex,
        month: SupportsIndex,
        day: SupportsIndex,
        hour: SupportsIndex = ...,
        minute: SupportsIndex = ...,
        second: SupportsIndex = ...,
        microsecond: SupportsIndex = ...,
        tzinfo: _MaybeTZ = None,
        *,
        fold: int = ...,
    ) -> DateTime[_MaybeTZ]: ...
    #
    # fix inherited stubs to consider generic timezone
    @property
    def tzinfo(self) -> _MaybeTZ: ...

    if sys.version_info >= (3, 12):
        @classmethod
        def fromtimestamp(
            cls, timestamp: float, tz: _OptionalTZ = None
        ) -> DateTime[_OptionalTZ]: ...
    else:
        @classmethod
        def fromtimestamp(
            cls, timestamp: float, /, tz: _OptionalTZ = None
        ) -> DateTime[_OptionalTZ]: ...

    @classmethod
    def utcfromtimestamp(cls, t: float, /) -> NaiveDateTime: ...
    #
    @classmethod
    def now(cls, tz: _OptionalTZ = None) -> DateTime[_OptionalTZ]: ...
    #
    @classmethod
    def utcnow(cls) -> NaiveDateTime: ...
    #
    @overload
    @classmethod
    def combine(
        cls, date: Date, time: Time[_OptionalTZ]
    ) -> DateTime[_OptionalTZ]: ...
    @overload
    @classmethod
    def combine(
        cls, date: Date, time: Time, tzinfo: _OptionalTZ
    ) -> DateTime[_OptionalTZ]: ...
    #
    def date(self) -> Date: ...
    def time(self) -> NaiveTime: ...
    def timetz(self) -> Time[_MaybeTZ]: ...

    if sys.version_info >= (3, 13):
        def __replace__(
            self,
            /,
            *,
            year: SupportsIndex = ...,
            month: SupportsIndex = ...,
            day: SupportsIndex = ...,
            hour: SupportsIndex = ...,
            minute: SupportsIndex = ...,
            second: SupportsIndex = ...,
            microsecond: SupportsIndex = ...,
            tzinfo: _OptionalTZ = None,
            fold: int = ...,
        ) -> DateTime[_OptionalTZ]: ...

    def replace(
        self,
        year: SupportsIndex = ...,
        month: SupportsIndex = ...,
        day: SupportsIndex = ...,
        hour: SupportsIndex = ...,
        minute: SupportsIndex = ...,
        second: SupportsIndex = ...,
        microsecond: SupportsIndex = ...,
        tzinfo: _OptionalTZ = None,
        *,
        fold: int = ...,
    ) -> DateTime[_OptionalTZ]: ...
    #
    @overload
    def astimezone(self, tz: None = None) -> AwareDateTime: ...
    @overload
    def astimezone(self, tz: _TZ) -> DateTime[_TZ]: ...
    #
    # prevent aware-vs-naive comparisons
    @overload
    def __le__(self: NaiveDateTime, value: NaiveDateTime, /) -> bool: ...
    @overload
    def __le__(self: AwareDateTime, value: AwareDateTime, /) -> bool: ...
    @overload
    def __lt__(self: NaiveDateTime, value: NaiveDateTime, /) -> bool: ...
    @overload
    def __lt__(self: AwareDateTime, value: AwareDateTime, /) -> bool: ...
    @overload
    def __ge__(self: NaiveDateTime, value: NaiveDateTime, /) -> bool: ...
    @overload
    def __ge__(self: AwareDateTime, value: AwareDateTime, /) -> bool: ...
    @overload
    def __gt__(self: NaiveDateTime, value: NaiveDateTime, /) -> bool: ...
    @overload
    def __gt__(self: AwareDateTime, value: AwareDateTime, /) -> bool: ...
    #
    @overload
    def __sub__(self: NaiveDateTime, value: NaiveDateTime, /) -> _timedelta: ...
    @overload
    def __sub__(self: AwareDateTime, value: AwareDateTime, /) -> _timedelta: ...
    @overload
    def __sub__(self, value: _timedelta, /) -> Self: ...
    #
    # fix stubs iherited from datetime.date that don't encode time
    @classmethod
    def today(cls) -> NaiveDateTime: ...
    @classmethod
    def fromordinal(cls, n: int, /) -> NaiveDateTime: ...
    @classmethod
    def fromisocalendar(
        cls, year: int, week: int, day: int
    ) -> NaiveDateTime: ...

NaiveDateTime: TypeAlias = DateTime[None]
"""Alias for DateTime with no timezone associated."""
AwareDateTime: TypeAlias = DateTime[_tzinfo]
"""Alias for DateTime with timezone associated."""

#
# === Utility functions ===

# typed()
@overload
def typed(dt: _time) -> Time: ...
@overload
def typed(dt: _datetime) -> DateTime: ...

# is_naive()
@overload
def is_naive(dt: _time) -> TypeIs[NaiveTime]: ...
@overload
def is_naive(dt: _datetime) -> TypeIs[NaiveDateTime]: ...

# is_aware()
@overload
def is_aware(dt: _time) -> TypeIs[AwareTime]: ...
@overload
def is_aware(dt: _datetime) -> TypeIs[AwareDateTime]: ...

# as_naive()
@overload
def as_naive(dt: _time) -> NaiveTime: ...
@overload
def as_naive(dt: _datetime) -> NaiveDateTime: ...

# as_aware()
@overload
def as_aware(dt: _time) -> AwareTime: ...
@overload
def as_aware(dt: _datetime) -> AwareDateTime: ...

# as_date()
def as_date(dt: _date) -> Date: ...



from datetime import timezone
from zoneinfo import ZoneInfo

import pytest

from datetypes import (
    AwareDateTime,
    AwareTime,
    Date,
    DateTime,
    NaiveDateTime,
    NaiveTime,
    Time,
    is_aware,
    is_naive,
)


def test_now():
    naive_dt = DateTime.now()
    aware_dt = DateTime.now(timezone.utc)
    enforced_naive_dt = DateTime.now(None)

    assert is_naive(naive_dt)
    assert is_aware(aware_dt)
    assert is_naive(enforced_naive_dt)


def test_combine():
    d = Date(2024, 1, 1)
    t = Time(12, 0)
    ttz = Time(12, 0, tzinfo=timezone.utc)

    dt_default_naive: NaiveDateTime = DateTime.combine(d, t)
    assert is_naive(dt_default_naive)

    dt_default_aware: AwareDateTime = DateTime.combine(d, ttz)
    assert is_aware(dt_default_aware)

    dt_tz_added_from_naive: AwareDateTime = DateTime.combine(d, t, timezone.utc)
    assert is_aware(dt_tz_added_from_naive)

    dt_tz_added_from_aware: AwareDateTime = DateTime.combine(
        d, ttz, timezone.utc
    )
    assert is_aware(dt_tz_added_from_aware)

    dt_tz_removed_from_naive: NaiveDateTime = DateTime.combine(d, t, None)
    assert is_naive(dt_tz_removed_from_naive)

    dt_tz_removed_from_aware: NaiveDateTime = DateTime.combine(d, ttz, None)
    assert is_naive(dt_tz_removed_from_aware)


def test_naive_components():
    naive_dt: DateTime = DateTime(2024, 1, 1, 12, 0)

    d: Date = naive_dt.date()
    assert isinstance(d, Date)

    t: NaiveTime = naive_dt.time()
    assert is_naive(t)

    ttz: NaiveTime = naive_dt.timetz()
    assert is_naive(ttz)


def test_aware_components():
    aware_dt: DateTime = DateTime(2024, 1, 1, 12, 0, tzinfo=timezone.utc)

    d: Date = aware_dt.date()
    assert isinstance(d, Date)

    t: NaiveTime = aware_dt.time()
    assert is_naive(t)

    ttz: AwareTime = aware_dt.timetz()
    assert is_aware(ttz)


@pytest.mark.parametrize(
    "dt",
    (
        DateTime(2024, 1, 1, 12, 0),
        DateTime(2024, 1, 1, 12, 0, tzinfo=timezone.utc),
    ),
    ids=("naive", "aware"),
)
def test_astimezone(dt: DateTime):
    local: AwareDateTime = dt.astimezone()
    assert is_aware(local)

    changed: AwareDateTime = dt.astimezone(timezone.utc)
    assert is_aware(changed)

    local_explicit: AwareDateTime = dt.astimezone(None)
    assert is_aware(local_explicit)


def test_comparisons():
    naive_dt: DateTime = DateTime(2024, 1, 1, 12, 0)
    aware_dt: DateTime = DateTime(2024, 1, 1, 12, 0, tzinfo=timezone.utc)
    aware_dt_zone: DateTime = DateTime(
        2024, 1, 1, 12, 0, tzinfo=ZoneInfo("UTC")
    )

    assert aware_dt >= aware_dt <= aware_dt
    assert aware_dt >= aware_dt_zone <= aware_dt
    assert naive_dt >= naive_dt <= naive_dt

    with pytest.raises(TypeError):
        assert aware_dt <= naive_dt  # pyright: ignore[reportOperatorIssue]
        assert naive_dt <= aware_dt  # pyright: ignore[reportOperatorIssue]

    with pytest.raises(TypeError):
        assert aware_dt_zone <= naive_dt  # pyright: ignore[reportOperatorIssue]
        assert naive_dt <= aware_dt_zone  # pyright: ignore[reportOperatorIssue]


def test_difference():
    naive_dt: DateTime = DateTime(2024, 1, 1, 12, 0)
    aware_dt: DateTime = DateTime(2024, 1, 1, 12, 0, tzinfo=timezone.utc)
    aware_dt_zone: DateTime = DateTime(
        2024, 1, 1, 12, 0, tzinfo=ZoneInfo("UTC")
    )

    assert (aware_dt - aware_dt).total_seconds() == 0
    assert (naive_dt - naive_dt).total_seconds() == 0
    assert (aware_dt - aware_dt_zone).total_seconds() == 0

    with pytest.raises(TypeError):
        assert aware_dt - naive_dt  # pyright: ignore[reportOperatorIssue]

    with pytest.raises(TypeError):
        assert naive_dt - aware_dt  # pyright: ignore[reportOperatorIssue]

    with pytest.raises(TypeError):
        assert naive_dt - aware_dt_zone  # pyright: ignore[reportOperatorIssue]


def test_inherited_from_date():
    ts: NaiveDateTime = DateTime.fromtimestamp(1234567890)
    assert isinstance(ts, DateTime)
    # assert not isinstance(ts, Date)

    today: NaiveDateTime = DateTime.today()
    assert isinstance(today, DateTime)
    # assert not isinstance(today, Date)

    ord: NaiveDateTime = DateTime.fromordinal(123)
    assert isinstance(ord, DateTime)
    # assert not isinstance(ord, Date)

    isoform: DateTime = DateTime.fromisoformat("2024-03-01")
    assert isinstance(isoform, DateTime)
    # assert not isinstance(isoform, Date)

    isocal: NaiveDateTime = DateTime.fromisocalendar(2024, 3, 2)
    assert isinstance(isocal, DateTime)
    # assert not isinstance(calendar, Date)



from datetime import date, datetime, time, timedelta, timezone

import pytest
from typing_extensions import assert_type

from datetypes import (
    AwareDateTime,
    AwareTime,
    Date,
    DateTime,
    NaiveDateTime,
    NaiveTime,
    Time,
    as_date,
    is_aware,
    is_naive,
    typed,
)


def native_function(
    d: date, t: time, dt: datetime
) -> tuple[date, time, datetime]:
    return d, t, dt


def typed_function(
    d: Date, t: Time, dt: DateTime
) -> tuple[Date, Time, DateTime]:
    return d, t, dt


def naive_function(t: NaiveTime, dt: NaiveDateTime) -> tuple[Time, DateTime]:
    return t, dt


def aware_function(t: AwareTime, dt: AwareDateTime) -> tuple[Time, DateTime]:
    return t, dt


def test_native_basic():
    d = date(2024, 1, 1)
    t = time(12, 0)
    dt = datetime(2024, 1, 2, 13, 0)

    # check function calls
    native_function(d, t, dt)
    typed_function(as_date(d), typed(t), typed(dt))

    # typing assertions
    assert_type(d, date)
    assert_type(t, time)
    assert_type(dt, datetime)

    # check native assertions
    assert isinstance(d, date)
    assert isinstance(t, time)
    assert isinstance(dt, datetime)

    # check typed assertions
    assert isinstance(d, Date)
    assert isinstance(t, Time)
    assert isinstance(dt, DateTime)


def test_typed_basic():
    d: Date = Date(2024, 1, 1)
    t: Time = Time(12, 0)
    dt: DateTime = DateTime(2024, 1, 2, 13, 0)

    # check function calls
    native_function(d, t, dt)
    typed_function(d, t, dt)

    # check native assertions
    assert isinstance(d, date)
    assert isinstance(t, time)
    assert isinstance(dt, datetime)

    # check typed assertions
    assert isinstance(d, Date)
    assert isinstance(t, Time)
    assert isinstance(dt, DateTime)


def test_naive_versions():
    t: Time = NaiveTime(12, 0)
    dt: DateTime = NaiveDateTime(2024, 1, 1, 15, 0)

    # check function calls
    naive_function(t, dt)
    native_function(dt.date(), t, dt)

    # check strict types # TODO
    # assert isinstance(t, NaiveTime)
    # assert isinstance(dt, NaiveDateTime)
    # assert not isinstance(t, AwareTime)
    # assert not isinstance(dt, AwareDateTime)

    # check lax types
    assert isinstance(t, Time)
    assert isinstance(dt, DateTime)

    # check native types
    assert isinstance(t, time)
    assert isinstance(dt, datetime)


def test_aware_versions():
    t: Time = AwareTime(12, 0, tzinfo=timezone.utc)
    dt: DateTime = AwareDateTime(2024, 1, 1, 15, 0, tzinfo=timezone.utc)

    # check function calls
    aware_function(t, dt)
    native_function(dt.date(), t, dt)

    # check strict types # TODO
    # assert isinstance(t, AwareTime)
    # assert isinstance(dt, AwareDateTime)
    # assert not isinstance(t, NaiveTime)
    # assert not isinstance(dt, NaiveDateTime)

    # check lax types
    assert isinstance(t, Time)
    assert isinstance(dt, DateTime)

    # check native types
    assert isinstance(t, time)
    assert isinstance(dt, datetime)


def test_incompatible_aliases():
    # no timezone provided, should type error
    actually_naive_t = AwareTime(12, 0)  # pyright: ignore[reportArgumentType]
    assert actually_naive_t

    actually_naive_dt = AwareDateTime(2024, 1, 1, 12, 0)  # pyright: ignore[reportArgumentType]
    assert actually_naive_dt

    # timezone provided, should type error
    actually_aware_t = NaiveTime(12, 0, tzinfo=timezone.utc)  # pyright: ignore[reportArgumentType]
    assert actually_aware_t

    actually_aware_dt = NaiveDateTime(2024, 1, 1, 12, 0, tzinfo=timezone.utc)  # pyright: ignore[reportArgumentType]
    assert actually_aware_dt


@pytest.mark.skip(reason="unclear whether we want generics at runtime?")
def test_generic_versions():
    aware_t: AwareTime = Time(12, 0, tzinfo=timezone.utc)
    assert isinstance(aware_t, AwareTime)  # pyright: ignore[reportArgumentType]

    aware_dt: AwareDateTime = DateTime(2024, 1, 1, 15, 0, tzinfo=timezone.utc)
    assert isinstance(aware_dt, AwareDateTime)  # pyright: ignore[reportArgumentType]

    naive_t: NaiveTime = Time(12, 0)
    assert isinstance(naive_t, NaiveTime)  # pyright: ignore[reportArgumentType]

    naive_dt: NaiveDateTime = DateTime(2024, 1, 1, 15, 0)
    assert isinstance(naive_dt, NaiveDateTime)  # pyright: ignore[reportArgumentType]


@pytest.mark.skip(reason="unclear whether we want generics at runtime?")
def test_generic_parameters():
    My_DateTime = DateTime[timezone]
    My_Time = Time[timezone]

    # just alias at runtime
    assert type(My_DateTime(2024, 1, 1, tzinfo=timezone.utc)) == datetime  # noqa: E721
    assert type(My_Time(12, 0, tzinfo=timezone.utc)) == time  # noqa: E721


def test_datetime_is_not_date():
    dt = DateTime(2024, 1, 1, 15, 0)
    assert_type(dt.date(), Date)

    # assert not isinstance(dt, Date) # TODO
    assert isinstance(dt.date(), Date)


def test_today():
    today = Date.today()
    assert_type(today, Date)


def test_replace():
    dt = DateTime(2024, 1, 1, 15, 0)
    assert_type(dt, NaiveDateTime)
    assert is_naive(dt)

    dt_with_tz = dt.replace(tzinfo=timezone.utc)
    assert_type(dt_with_tz, "DateTime[timezone]")
    assert is_aware(dt_with_tz)

    dt_with_tz_removed = dt_with_tz.replace(tzinfo=None)
    assert_type(dt_with_tz_removed, NaiveDateTime)
    assert is_naive(dt_with_tz_removed)

    t = Time(13, 0)
    assert_type(t, NaiveTime)
    assert is_naive(t)

    t_with_tz = t.replace(tzinfo=timezone.utc)
    assert_type(t_with_tz, "Time[timezone]")
    assert is_aware(t_with_tz)

    t_with_tz_removed = t.replace(tzinfo=None)
    assert_type(t_with_tz_removed, NaiveTime)
    assert is_naive(t_with_tz_removed)


def test_maths():
    d = Date(2024, 1, 1)
    t = Time(15, 0)
    dt = DateTime.combine(d, t)

    d_plus_one: Date = d + timedelta(days=1)
    assert d_plus_one

    d_minus_one: Date = d - timedelta(days=1)
    assert d_minus_one

    dt_plus_one: DateTime = dt + timedelta(days=1)
    assert dt_plus_one

    dt_minus_one: DateTime = dt - timedelta(days=1)
    assert dt_minus_one

    diff1: timedelta = d - d
    assert diff1.total_seconds() == 0

    diff2: timedelta = dt - dt
    assert diff2.total_seconds() == 0

    with pytest.raises(TypeError):
        assert d - dt  # pyright: ignore[reportOperatorIssue]

    with pytest.raises(TypeError):
        assert dt - d  # pyright: ignore[reportOperatorIssue]



# type: ignore

from datetime import date as _date
from datetime import datetime as _datetime
from datetime import time as _time

# --- basic symbols ---

Date = _date
Time = _time
DateTime = _datetime

# --- naive vs aware ---
try:
    # attach annotate-types for runtime introspection, if available
    from typing import Annotated

    from annotated_types import (
        Timezone,
    )

    NaiveTime = Annotated[Time, Timezone(None)]
    AwareTime = Annotated[Time, Timezone(...)]

    NaiveDateTime = Annotated[DateTime, Timezone(None)]
    AwareDateTime = Annotated[DateTime, Timezone(...)]

except Exception:
    # just alias symbols otherwise
    NaiveTime = Time
    AwareTime = Time

    NaiveDateTime = DateTime
    AwareDateTime = DateTime

# attach support for generic syntax at runtime with minimal work.
#
# NOTE: This doesn't seem to work - I don't want to resort to an actual subclass
# using Generic (as done in the stubs), even if I were to return the concrete
# type on __new__(), otherwise the type may not play well with Pydantic and
# other libraries relying on runtime type annotations.
#
# Time.__class_getitem__ = lambda cls, _: cls
# DateTime.__class_getitem__ = lambda cls, _: cls


# --- utility functions ---


def typed(dt):
    return dt


def is_naive(dt):
    # see: https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive
    arg = dt if isinstance(dt, _datetime) else None
    return dt.tzinfo is None or dt.tzinfo.utcoffset(arg) is None


def is_aware(dt):
    # see: https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive
    arg = dt if isinstance(dt, _datetime) else None
    return dt.tzinfo is not None and dt.tzinfo.utcoffset(arg) is not None


def as_naive(dt):
    assert is_naive(dt), f"Expected naive object, received {dt!r}"
    return dt


def as_aware(dt):
    assert is_aware(dt), f"Expected aware object, received {dt!r}"
    return dt


def as_date(dt):
    assert type(dt) is _date, f"Excepted datetime.date object, received {dt!r}"
    return dt
